# -*- coding: japanese-shift-jis-dos; tab-width: 4; -*-
#
# wdic-menu - FreePWING script for WDIC 2001
# $Id: wdic-menu.pl 1.3.0.1 2002/08/02 05:45:12 satomii Exp $
#
# This file is designed to be used from the wdic-fpw.pl script.
# DO NOT CHANGE THE ENCODING OF THIS FILE.
#
# Copyright (C) 2000-2002, rei@wdic.org.
# This program is distributed in the hope that it will be useful, but
# without any warranty. See the GNU General Public License for the details.
#
require 5.005;

use strict;
use File::DosGlob 'glob';
use IO::File;
use Jcode;

my @copyright;
my $index_title = 'O[vʎ^ꗗ';
my $plugin_title = '^vOCꗗ';

#
# void make_menu(void)
# writes the description about each groups in the current text context.
#
sub make_menu()
{
	my(@handle, @list, $title, $index, $str, $prev_toc, $next_toc);
	my $menu_tag = 'WDIC_MAN_%02X';

	print "Writing menu and copyright pages...\n";

	##
	## the top menu page
	##
	# readme.txt, required. the file existence check should have been done
	# in wdic-fpw.pl, but we use glob again to ignore case (if possible
	# and applicable).
	$list[0] = glob("./readme.txt");
	$list[0] = "./readme.txt" if (!$list[0]);
	$handle[0] = new IO::File($list[0], 'r');
	die "Unable to open file: ./readme.txt ($!)" if (!$handle[0]);
	binmode $handle[0];
	while ($_ = $handle[0]->getline()) {
		if (s/^([^\r\n]+[\r\n]+)$/$1/) {
			&write_link($fpwmenu, $_, 'README');
			last;
		}
	}

	# WDIC.MAN, required.
	$list[0] = glob("$wdic_dir/WDIC.MAN");
	$list[0] = "$wdic_dir/WDIC.MAN" if (!$list[0]);
	$handle[1] = new IO::File($list[0], 'r');
	die "Unable to open file: $wdic_dir/WDIC.MAN ($!)" if (!$handle[1]);
	binmode $handle[1];
	while ($_ = $handle[1]->getline()) {
		if (s/^\s*(ʐMp̊bm)\s*([0-9]+[\r\n]+)$/$1 $2/) {
			&write_link($fpwmenu, $_, 'WDIC_MAN');
			$title = $_;
			last;
		}
	}

	# QUESTION.TXT, optional.
	$list[0] = glob("$wdic_dir/QUESTION.TXT");
	$list[0] = "$wdic_dir/QUESTION.TXT" if (!$list[0]);
	$handle[2] = new IO::File($list[0], 'r');
	if (defined($handle[2])) {
		binmode $handle[2];
		while ($_ = $handle[2]->getline()) {
			if (s/^\s+([^\r\n]+[\r\n]+)$/$1/) {
				&write_link($fpwmenu, $_, 'QUESTION');
				last;
			}
		}
	}

	# index by group, if tag files are found.
	@list = glob("$wdic_dir/*.TAG");
	&write_link($fpwmenu, "$index_title\n", 'INDEX') if (@list);

	# a list of plugins, if --cgraphs and/or --plugin is set.
	# to generate the actual entry, make_plugin_list must be called later.
	if ($opt_cgraphs || $opt_plugin) {
		&write_link($fpwmenu, "$plugin_title\n", 'PLUGIN');
	}

	##
	## readme.txt
	##
	print "readme.txt\n";
	$handle[0]->seek(0, IO::File::SEEK_SET);
	while ($_ = $handle[0]->getline()) {
		last if (s/^([^\r\n]+)[\r\n]+$/$1/);
	}
	&new_menu_context($_, 'README');
	while ($_ = $handle[0]->getline()) {
		&write_sjis($fpwmenu, $_);
	}
	$handle[0]->close();

	##
	## WDIC.MAN
	##
	print "WDIC.MAN\n";
	&new_menu_context('', 'WDIC_MAN');

	$index = 0;
	while ($_ = $handle[1]->getline()) {
		if (s/^\-+  (.+)  \-+[\r\n]+$/$1/) {
			$index++;
			$_ = &reform_sjis($_, 1) if (/\x81[\x69\x6a]/);
			&write_link($fpwmenu, "$_\n", sprintf($menu_tag, $index));
		}
	}

	$index = 0;
	$handle[1]->seek(0, IO::File::SEEK_SET);
	while ($str = $handle[1]->getline()) {
		last if ($str =~ s/^\-+  (.+)  \-+[\r\n]+$/$1/);
	}

	do {
		print "  $str\n";
		$index++;
		&new_menu_context($str, sprintf($menu_tag, $index));

		# dispatch the section to the writer-function.
		if ('e̓F' eq $str) {
			$next_toc = &write_menu_section_features($handle[1]);
		} elsif ('tH[}bg̑' eq $str) {
			$next_toc = &write_menu_section_format($handle[1]);
		} elsif ('iɂ' eq $str) {
			$next_toc = &write_menu_section_pos($handle[1]);
		} elsif ('ǂ݂ɂ' eq $str || 'ɂ' eq $str) {
			$next_toc = &write_menu_section_reading($handle[1]);
		} elsif ('̑̋K' eq $str) {
			$next_toc = &write_menu_section_rules($handle[1]);
		} elsif ('b' eq $str) {
			$next_toc = &write_menu_section_references($handle[1]);
		} elsif ('ₗ' eq $str || '' eq $str) {
			$next_toc = &write_menu_section_versions($handle[1]);
		} elsif ('M' eq $str) {
			$next_toc = &write_menu_section_environment($handle[1]);
		} else {
			# 1.3: the rest is not (yet) separately handled.
			$next_toc = &write_menu_section($str, $handle[1]);
		}
		$next_toc = &reform_sjis($_, 1) if ($next_toc =~ /\x81[\x69\x6a]/);

		# menu footer (navigation).
		$fpwmenu->add_indent_level(2);
		&write_newline($fpwmenu, 1);
		if ($prev_toc) {
			&write_sjis($fpwmenu, '');
			&write_link($fpwmenu, $prev_toc, sprintf($menu_tag, $index - 1));
		}
		if ($next_toc) {
			&write_sjis($fpwmenu, ($prev_toc ? ', ' : '') . '');
			&write_link($fpwmenu, $next_toc, sprintf($menu_tag, $index + 1));
		}
		&write_sjis($fpwmenu, ', ');
		&write_link($fpwmenu, "$title\n", 'WDIC_MAN');
		$prev_toc = $str;
		$str = $next_toc;
	} while ($_);

	$handle[1]->close();

	##
	## QUESTION.TXT (if any)
	##
	if (defined($handle[2])) {
		print "QUESTION.TXT\n";
		$handle[2]->seek(0, IO::File::SEEK_SET);
		while ($_ = $handle[2]->getline()) {
			last if (s/^\s+([^\r\n]+)[\r\n]+$/$1/);
		}
		&new_menu_context($_, 'QUESTION');
		while ($_ = $handle[2]->getline()) {
			if (/^\- \- \-/) {
				&write_sjis($fpwmenu, $_);
				last;
			}
			s/^ {5}//;
			s/([^\r\n]*)[\r\n]+$/$1/ if (!/^ / && !/B\r\n$/);
			&write_sjis($fpwmenu, $_);
		}
		while ($_ = $handle[2]->getline()) {
			&write_sjis($fpwmenu, $_);
			last if (/^\- \- \-/);
		}
		while ($_ = $handle[2]->getline()) {
			s/^ {5}//;
			&write_sjis($fpwmenu, $_)
		}
		$handle[2]->close();
	}

	##
	## index by group
	##
	&make_index(@list) if (@list);
}

#
# void new_menu_context(char *title, char *tag)
# creates a new menu context with the given title and tag. the indentation
# of the context is set to 2 by default.
#
sub new_menu_context($$)
{
	my($title, $tag) = @_;

	$fpwmenu->new_context()
		|| die &death_cry("Unable to add menu context", $fpwmenu);
	if ($fpwmenu->can('add_entry_tag')) {
		$fpwmenu->add_entry_tag($tag)
			|| die &death_cry("Unable to add entry tag: $tag", $fpwmenu);
	} else {
		$fpwmenu->add_tag($tag)
			|| die &death_cry("Unable to add entry tag: $tag", $fpwmenu);
	}
	return if (!$title);

	if (!$fpwmenu->add_keyword_start()
			|| !$fpwmenu->add_text(Jcode::convert($title, 'euc', 'sjis'))
			|| !$fpwmenu->add_keyword_end()
			|| !$fpwmenu->add_newline()
			|| !$fpwmenu->add_indent_level(2)) {
		die &death_cry("Unable to write menu title", $fpwmenu);
	}
}

##-------------------------------------------------------------------------
## MENU WRITER FUNCTIONS

#
# writes the "kaku-jisho no tokushoku" section into the current menu
# context, then returns the first line of the next section.
#
sub write_menu_section_features($)
{
	my ($handle) = @_;
	my ($newline, $indent, $prev) = (1, 2, '');
	my ($file, $cat);

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		last if (s/^\-+  (.+)  \-+$/$1/ || s/^\-$//);

		if (/^ *$/) {
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}

		} elsif (s/^ +(\[[A-Z]+ *$SJIS_ANY+\])/$1/) {
			if (2 != $indent) {
				$fpwmenu->add_indent_level(2);
				$indent = 2;
			}
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 2);

		} elsif (s/^ +([A-Z]+\.DIC) +([^ ]+) +//) {
			$file = $1 . (" " x (15 - length($1)));
			$cat = $2 . (" " x (20 - length($2)));
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}
			$fpwmenu->add_nowrap_start();
			&write_sjis($fpwmenu, "$file$cat" . &reform_sjis($_, 1));
			$fpwmenu->add_nowrap_end();
			$newline = &write_newline($fpwmenu, 1);

		} elsif (s/^ +([A-Z]+\.[A-Z]+) +([^ ])/$2/) {
			$file = $1 . (" " x (15 - length($1)));
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}
			$fpwmenu->add_nowrap_start();
			&write_sjis($fpwmenu, $file . &reform_sjis($_, 1));
			$fpwmenu->add_nowrap_end();
			$newline = &write_newline($fpwmenu, 1);

		} elsif (3 == $indent) {
			s/^ +//;
			$fpwmenu->add_indent_level(2);
			$indent = 2;
			&write_newline($fpwmenu, 2 - $newline) if ($newline < 2);
			$prev = &reform_sjis($_, 1);
			&write_sjis($fpwmenu, $prev);
			if (length($prev) < 15 && $prev =~ /^$SJIS_ANY+$/) {
				$newline = &write_newline($fpwmenu, 2);
			} else {
				$newline = 0;
				next;
			}

		} else {
			s/^ +//;
			$_ = &reform_sjis($_, 1);
			s/ +$//;
			if ($prev && ($prev !~ /$SJIS_ANY$/ || !/^$SJIS_ANY/)) {
				$fpwmenu->add_text(" ");
			}
			&write_sjis($fpwmenu, $_);
			$prev = $_;
			$newline = 0;
			next;
		}

		$prev = '';
	}

	return $_;
}

#
# writes the "format sono-ta" section into the current menu context, then
# returns the first line of the next section.
#
sub write_menu_section_format($)
{
	my ($handle) = @_;
	my ($newline, $prev) = (1, '');

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		last if (s/^\-+  (.+)  \-+$/$1/ || s/^\-$//);

		if (/^ *$/) {
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}

		} elsif (s/^ {3}\x81[\xa1\x45] *//) {
			&write_sjis($fpwmenu, " $_");
			$newline = &write_newline($fpwmenu, 2);

		} elsif (s/^ {6} *(#|y$SJIS_ANY+z)//) {
			my $prefix = $1;
			if ('#' ne $prefix) {
				$prefix = &reform_sjis($prefix, 0);
				$_ = &reform_sjis($_, 0);
				s/^(>+) +/$1/;
			}
			$fpwmenu->add_indent_level(3);
			$fpwmenu->add_nowrap_start();
			&write_sjis($fpwmenu, "$prefix$_");
			$fpwmenu->add_nowrap_end();
			&write_newline($fpwmenu, 1);
			while ($_ = $handle->getline()) {
				last if (/^[ \r\n]*$/);
				s/[ \r\n]+$//;
				s/^ +//;
				if ('#' eq $prefix) {
					$_ = (" " x 8) . &reform_sjis($_, 1);
				} else {
					$_ = &reform_sjis($_, 0);
					s/^(>+) +/$1/;
					$_ = (" " x length($prefix)) . $_;
				}
				$fpwmenu->add_nowrap_start();
				&write_sjis($fpwmenu, $_);
				$fpwmenu->add_nowrap_end();
				&write_newline($fpwmenu, 1);
			}
			$fpwmenu->add_indent_level(2);
			&write_newline($fpwmenu, 1);
			$newline = 2;

		} else {
			s/^ +//;
			$_ = &reform_sjis($_, 1);
			s/ +$//;
			if ($prev && ($prev !~ /$SJIS_ANY$/ || !/^$SJIS_ANY/)) {
				$fpwmenu->add_text(" ");
			}
			&write_sjis($fpwmenu, $_);
			$prev = $_;
			$newline = 0;
			next;
		}

		$prev = '';
	}

	return $_;
}

#
# writes the "hinshi-mei ni tsuite" section into the current menu context,
# then returns the first line of the next section.
#
sub write_menu_section_pos($)
{
	my ($handle) = @_;
	my ($newline, $indent, $prev) = (1, 2, '');

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		last if (s/^\-+  (.+)  \-+$/$1/ || s/^\-$//);

		if (/^ *$/) {
			$newline = &write_newline($fpwmenu, 1) if (!$newline);
			if (2 != $indent) {
				$fpwmenu->add_indent_level(2);
				$indent = 2;
			}
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}
			$prev = '';
			next;

		} elsif ($prev) {
			s/^ +//;
			$_ = &reform_sjis($_, 1);
			s/ +$//;
			if ($prev !~ /$SJIS_ANY$/ || !/^$SJIS_ANY/) {
				$fpwmenu->add_text(" ");
			}
			&write_sjis($fpwmenu, $_);
			$prev = $_;
			$newline = 0;
			next;

		} elsif (s/^ + *//) {
			&write_sjis($fpwmenu, " $_");
			$newline = &write_newline($fpwmenu, 2);
			$prev = '';
			next;

		} elsif (s/^ +()/$1/) {
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 1);
			$prev = '';
			next;

		} elsif (s/^  (\x84[\xab\xac\xaf])/$1/) {
			# table, nop here.

		} elsif (s/^ {6}((?:[ \|]+)?\+\@[^ ]+ *)//) {
			# semantic part-of-speech tree.
			my $name = $1;
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}
			if ($_) {
				$_ = &reform_sjis($_, 1);
				s/(\@$SJIS_ANY+) (ETC)/$1$2/g;
			}
			$_ = "$name$_";

		} elsif (s/^ {6}([ \|]+)$/$1/) {
			# semantic part-of-speech node, branch only. nop.

		} elsif (s/^ {6}( +\+[^ ]+)//) {
			# child node of extended part-of-speech.
			my $name = $1;
			$_ = &reform_sjis($_, 1) if ($_);
			$_ = $name . (' ' x (20 - length($name))) . $_;

		} elsif (s/^ {5} ?($SJIS_ANY+)$/$1/ && length($_) <= 10) {
			# parent node of extended part-of-speech.
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}

		} elsif (s/^ {8} *($SJIS_ANY+  +[\@\+])/$1/) {
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}

		} else {
			s/^ +//;
			$prev = &reform_sjis($_, 1);
			$prev =~ s/ +$//;
			&write_sjis($fpwmenu, $prev);
			$newline = 0;
			next;
		}

		$fpwmenu->add_nowrap_start();
		&write_sjis($fpwmenu, $_);
		$fpwmenu->add_nowrap_end();
		$newline = &write_newline($fpwmenu, 1);
		$prev = '';
	}

	return $_;
}

#
# writes the "yomi ni tsuite" or "hatsuon ni tsuite" section into the
# current menu context, then returns the first line of the next section.
#
sub write_menu_section_reading($)
{
	my ($handle) = @_;
	my ($newline, $indent, $prev) = (1, 2, '');

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		last if (s/^\-+  (.+)  \-+$/$1/ || s/^\-$//);

		if (/^ *$/) {
			$newline = &write_newline($fpwmenu, 1) if (!$newline);
			if (2 != $indent) {
				$fpwmenu->add_indent_level(2);
				$indent = 2;
			}
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}
			$prev = '';
			next;

		} elsif ($prev) {
			s/^ +//;
			$_ = &reform_sjis($_, 1);
			s/ +$//;
			if ($prev !~ /$SJIS_ANY$/ || !/^$SJIS_ANY/) {
				$fpwmenu->add_text(' ');
			}
			&write_sjis($fpwmenu, $_);
			$prev = $_;
			$newline = 0;
			next;

		} elsif (s/^ {3}E/ / || s/^ {4}E($SJIS_ANY+)$/ $1/) {
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 2);
			$prev = '';
			next;

		} elsif (s/^ {10} *(E$SJIS_ANY+)$/$1/) {
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}
			$_ = &reform_sjis($_, 1);

		} elsif (s/^ {8} *(\[\])/$1/
				|| s/^ {8} ?([^ ]+  )/$1/
				|| s/^ {6}( *[^ ]+  )/$1/
				|| s/^ {5}([^ ].*  )/$1/) {
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}

		} else {
			s/^ +//;
			s/^E//;
			$prev = &reform_sjis($_, 1);
			$prev =~ s/ +$//;
			&write_sjis($fpwmenu, $prev);
			$newline = 0;
			next;
		}

		$fpwmenu->add_nowrap_start();
		&write_sjis($fpwmenu, $_);
		$fpwmenu->add_nowrap_end();
		$newline = &write_newline($fpwmenu, 1);
		$prev = '';
	}

	return $_;
}

#
# writes the "sono-ta no kitei" section into the current menu context, then
# returns the first line of the next section.
#
sub write_menu_section_rules($)
{
	my ($handle) = @_;
	my ($newline, $indent, $prev) = (1, 2, '');

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		last if (s/^\-+  (.+)  \-+$/$1/ || s/^\-$//);

		if (/^ *$/) {
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}

		} elsif (s/^ E// || s/^  +E/ /) {
			# the first line or heading.
			if (2 != $indent) {
				$fpwmenu->add_indent_level(2);
				$indent = 2;
			}
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 1);

		} else {
			s/^ +//;
			if (/  / || (!/,/ && !/\.$/)) {
				if (s/^((?:\x82[\x4f-\x58])+) +//) {
					my $mon = &reform_sjis($1, 1);
					$_ = $mon . (' ' x (13 - (length($mon)))) . $_;
				} elsif (s/^(\x81.) +//) {
					$_ = $1 . (' ' x 11) . &reform_sjis($_, 1);
				}
				if (3 != $indent) {
					$fpwmenu->add_indent_level(3);
					$indent = 3;
				}
				$fpwmenu->add_nowrap_start();
				&write_sjis($fpwmenu, $_);
				$fpwmenu->add_nowrap_end();
				$newline = &write_newline($fpwmenu, 1);
			} else {
				$_ = &reform_sjis($_, 1);
				s/ +$//;
				if ($prev) {
					if ($prev !~ /$SJIS_ANY$/ || !/^$SJIS_ANY/) {
						$fpwmenu->add_text(" ");
					}
				} elsif (2 != $indent) {
					$fpwmenu->add_indent_level(2);
					$indent = 2;
				}
				&write_sjis($fpwmenu, $_);
				$prev = $_;
				$newline = 0;
				next;
			}
		}
		$prev = '';
	}

	return $_;
}

#
# writes the "kiso bunken" section into the current menu context, then
# returns the first line of the next section.
#
sub write_menu_section_references($)
{
	my ($handle) = @_;
	my ($newline, $prev) = (1, '');

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		last if (s/^\-+  (.+)  \-+$/$1/ || s/^\-$//);

		if (/^ *$/) {
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}
			$prev = '';

		} elsif (s/^[\- ]+ +([^ ].*[^ ]) +[\- ]+$/ $1/) {
			$_ = &reform_sjis($_, 1);
			&write_sjis($fpwmenu, $_);
			$newline = &write_newline($fpwmenu, 2);
			$prev = '';

		} elsif (s/^  (RFC [0-9]+ +)//) {
			$prev = "$1: " . &reform_sjis($_, 1);
			&write_newline($fpwmenu, 1) if (!$newline);
			&write_sjis($fpwmenu, $prev);
			$newline = 0;

		} elsif (s/^ +(E)/$1/) {
			s/((?: |$SJIS_ANY)\x81\x5e)([^ ])/$1 $2/g;
			s/  +/ /g;
			&write_newline($fpwmenu, 1) if (!$newline);
			$prev = &reform_sjis($_, 1);
			$prev =~ s/ +$//;
			&write_sjis($fpwmenu, $prev);
			$newline = 0;

		} elsif (/^ *(.+)$/) {
			&write_newline($fpwmenu, 2 - $newline) if ($newline < 2);
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 1);
			$prev = '';

		} elsif (s/^k(.+)l$/$1/) {
			&write_newline($fpwmenu, 2 - $newline) if ($newline < 2);
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 1);
			$prev = '';

		} elsif (!$prev && (4 <= length($_) || /^$SJIS_ANY+n$/)) {
			s/^ +//;
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 2);

		} else {
			s/( \x81\x5e)([^ ])/$1 $2/g;
			s/^ +//;
			$_ = &reform_sjis($_, 1);
			s/ +$//;
			if ($prev && ($prev !~ /$SJIS_ANY$/ || !/^$SJIS_ANY/)) {
				$fpwmenu->add_text(" ");
			}
			&write_sjis($fpwmenu, $_);
			$prev = $_;
			$newline = 0;
		}
	}

	return $_;
}

#
# writes the "kaitei rireki" section into the current menu context, then
# returns the first line of the next section.
#
sub write_menu_section_versions($)
{
	my ($handle) = @_;
	my ($next, $space, $max, @table, @comment);

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;
		if (s/^\-+  (.+)  \-+$/$1/ || s/^\-$//) {
			$next = $_;
			last;
		} elsif (s/^ *//) {
			push(@comment, $_);
		} elsif (/^( +)-+[ \\-]+$/) {
			$space = length($1) if (!$space);
			push(@table, $_);
		} elsif (0 < scalar(@comment)) {
			s/^ +//;
			if (/^ *$/) {
				push(@comment, $_) if ($comment[$#comment] !~ /^$/);
			} else {
				$comment[$#comment] .= $_;
			}
		} else {
			if (/^ *$/) {
				push(@table, '') if (!@table || $table[$#table] !~ /^$/);
			} else {
				if (s/( [0-9,]+ +)([^0-9, ].+)$/$1/) {
					my $str = &reform_sjis($2, 1);
					$_ .= $str;
					$max = length($str) if ($max < length($str));
				}
				push(@table, $_);
			}
		}
	}

	foreach (@table) {
		s/^ {$space}//;
		if ($_) {
			$fpwmenu->add_nowrap_start();
			if (s/^([\-= ]+ )([\-=])\2+$/$1/) {
				$fpwmenu->add_text($_ . ($2 x $max))
					|| die &death_cry("Unable to write text", $fpwmenu);
			} else {
				&write_sjis($fpwmenu, $_);
			}
			$fpwmenu->add_nowrap_end();
		}
		&write_newline($fpwmenu, 1);
	}

	foreach (@comment) {
		&write_sjis($fpwmenu, &reform_sjis($_, 1)) if ($_);
		&write_newline($fpwmenu, 1);
	}

	return $next;
}

#
# writes the "shippitsu kankyou" section into the current menu context,
# then returns the first line of the next section.
#
sub write_menu_section_environment($)
{
	my ($handle) = @_;
	my ($newline, $indent, $prev) = (1, 2, '');
	my $name;

	while ($_ = $handle->getline()) {
		s/[ \r\n]+$//;

		if (s/^\-+  (.+)  \-+$/$1/ || s/^\-$//) {
			last;

		} elsif (/^ *$/) {
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}

		} elsif (s/^ {5}(?:  )?((?:[^ ]+ ?)+) {4} *([^ ])/$2/
				 || s/^ {20} *(ver)/$1/i) {
			if ($2) {
				$name = &reform_sjis($1, 1);
				$name .= (" " x (24 - length($name)));
			} else {
				$name = " " x 24;
			}
			if (s/(  )((?:[^ ]+ ?)+)$/$1/) {
				$_ = $_ . &reform_sjis($2, 1);
			}
			if (3 != $indent) {
				$fpwmenu->add_indent_level(3);
				$indent = 3;
			}
			$fpwmenu->add_nowrap_start();
			&write_sjis($fpwmenu, "$name$_");
			$fpwmenu->add_nowrap_end();
			$newline = &write_newline($fpwmenu, 1);

		} elsif (s/^ {5}([^ ])/ $1/) {
			if (2 != $indent) {
				$fpwmenu->add_indent_level(2);
				$indent = 2;
			}
			&write_newline($fpwmenu, 2 - $newline) if ($newline < 2);
			&write_sjis($fpwmenu, &reform_sjis($_, 1));
			$newline = &write_newline($fpwmenu, 2);

		} else {
			if (2 != $indent) {
				$fpwmenu->add_indent_level(2);
				$indent = 2;
			}
			s/^ +//;
			if ('fXNgbv' eq $_ || 'm[g' eq $_) {
				$prev = " $_";
			} else {
				$_ = &reform_sjis($_, 1);
				s/ +$//;
				if ($prev && ($prev !~ /$SJIS_ANY$/ || !/^$SJIS_ANY/)) {
					$fpwmenu->add_text(" ");
				}
				$prev = $_;
			}
			&write_sjis($fpwmenu, $prev);
			$newline = 0;
			next;
		}

		$prev = '';
	}

	return $_;
}

#
# char *write_menu_section(char *title, $handle)
# converts a WDIC.MAN section into a menu entry.
#
sub write_menu_section($$)
{
	my($title, $handle) = @_;
	my($newline, $indent, $prev_indent, $prev_str);

	$newline = 1;
	$indent = $prev_indent = 2;
	$prev_str = '';

	while ($_ = $handle->getline()) {
		if (s/^\-+  (.+)  \-+[\r\n]+$/$1/
				|| s/^\-+[\r\n]+$//) {
			last;

		} elsif (/^[\r\n]+$/) {
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}
			$prev_str = '';
			next;

		} elsif ('T|[g' eq $title) {
			if (s/^ *(\(\x82[\x4f-\x58]\)) */$1 /) {
				$_ = &reform_sjis($_, 1);
				$indent = 2;
			} else {
				s/^ +//;
				s/([^\r\n])[\r\n]+$/$1/ if (!/^[\x20-\x7e]+$/);
				$indent = 3;
			}

		} elsif ('t' eq $title || 'A' eq $title) {
			s/^ +//;
			push(@copyright, $_);

		} elsif (s/^ *E// || s/^ *(\(\%[0-9]\) )/$1/) {
			# a line with bullet or notation index.
			if ($newline < 2) {
				&write_newline($fpwmenu, 2 - $newline);
				$newline = 2;
			}
			$prev_str = '';
			s/([^\r\n]*)[\r\n]+$/$1/;
			$indent = 2;

		} elsif (s/^ *([0-9]\. *)/$1 /) {
			# also with an outline number.
			$newline = &write_newline($fpwmenu, 1) if (!$newline);
			$prev_str = '';
			s/([^\r\n]*)[\r\n]+$/$1/;

		} elsif (s/^ *([a-z]+:\/\/[\x21-\x7e]+[\r\n]+)$/  $1/) {
			# URL-only line.
			$indent = 3;

		} else {
			# default processing.
			s/^ +//;
			s/([^\r\n])[\r\n]+$/$1/;
			$indent = 2;
		}

		# write the current string to the menu context.
		if ($indent != $prev_indent) {
			$fpwmenu->add_indent_level($indent)
				|| die &death_cry("Unable to set indent level ($indent)",
					$fpwmenu);
			$prev_indent = $indent;
		}
		$newline = /[\r\n]+$/;
		if ($newline) {
			$prev_str = '';
		} else {
			$_ = &reform_sjis($_, 1);
			if ('' ne $prev_str && $prev_str !~ / $/ && !/^ /) {
				if ($prev_str !~ /$SJIS_ANY$/ || /^[\x21-\x7e]/) {
					$_ = " $_";
				}
			}
			$prev_str = $_;
		}
		&write_sjis($fpwmenu, $_);

		# more fix by my preference.
		if (/^/) {
			while ($newline < 2) {
				$fpwmenu->add_newline()
					|| die &death_cry("Failed to add newline", $fpwmenu);
				$newline++;
			}
		}
	}

	return $_;
}

##-------------------------------------------------------------------------
## INDEX

#
# void make_index(char **file)
# reads *.TAG files and converts them into index entries.
#
sub make_index(@)
{
	my (@list, %group, $i, $key, $cat);

	print "TAG files were found, creating group index...\n";

	@list = $wdic->get_group_list();
	for ($i = 0; $i < scalar(@list); $i++) {
		$group{$list[$i]} = $i;
	}
	@list = sort {
		my $a1 = uc($a);
		my $b1 = uc($b);
		$a1 =~ s/^.*\/([^\/]+)\.TAG/$1/;
		$b1 =~ s/^.*\/([^\/]+)\.TAG/$1/;
		if (defined($group{$a1})) {
			defined($group{$b1}) ? $group{$a1} <=> $group{$b1} : 1;
		} elsif (defined($group{b1})) {
			-1;
		} else {
			$a1 cmp $b1;
		}
	} @_;

	undef %group;
	&new_menu_context('', 'INDEX');

	# write the links to each group.
	foreach (@list) {
		my %hash = (
			HANDLE => new IO::File($_, 'r'),
			TAG => '',
			NAME => ''
		);

		die "Unable to open '$_' ($!)" if (!defined($hash{HANDLE}));
		binmode $hash{HANDLE};

		$_ = uc($_);
		s/^.*\/([^\/]+)\.TAG/$1/;

		$hash{TAG} = $_ . "_TAG";
		$hash{NAME} = $wdic->get_group_title($_);
		$group{$_} = \%hash;

		&write_link($fpwmenu, "$hash{NAME}\n", $hash{TAG});
	}

	# list up the words in the .TAG file.
	foreach $key (keys(%group)) {
		my %hash = %{ $group{$key} };
		print "$key.TAG: $hash{NAME}\n ";

		# create heading index.
		&new_menu_context('', $hash{TAG});
		&write_link($fpwmenu, "($index_title)\n", 'INDEX');

		while ($_ = $hash{HANDLE}->getline()) {
			last if (/^\-+[\r\n]+$/);
			next if (!s/^[A-Z]+\.TAG\([0-9]+\): +y (.+) z.*$/$1/);
			if ('`' eq $_) {
				$_ = 'At@xbg';
			} elsif (/^\x82[\xa0\xa9\xb3\xbd\xc8\xcd\xdc\xe2\xe7]$/) {
				# [Ȃ͂܂]
				$_ .= 's';
			} elsif ('' eq $_) {
				$_ .= 's, ';
			} elsif (length($_) < 4) {
				next;
			}
			&write_link($fpwmenu, "$_\n",
					sprintf("%s%04X", $hash{TAG}, unpack('S', $_)));
		}

		# create word index.
		while ($_ = $hash{HANDLE}->getline()) {
			s/([^\r\n]*)[\r\n]+$/$1/;
			next if (/^\-*$/);

			if (s/^\-+  (.+)  \-+$/$1/) {
				# heading line.
				print $_;
				$newline_on_death = 1;
				if (/^\x82[\x60\xa0\xa9\xb3\xbd\xc8\xcd\xdc\xe2\xe7\xed]$/
						|| 4 <= length($_)) {
					# [`Ȃ͂܂], L, 
					$_ = 'A' if (/^`$/);
					&new_menu_context('',
							sprintf("%s%04X", $hash{TAG}, unpack('S', $_)));
					&write_link($fpwmenu, "($hash{NAME})\n", $hash{TAG});
				}
				next;
			} elsif (!s/^([A-Z]+)\.DIC\([0-9]+\): +\#(.+)$/$2/) {
				next;
			}
			undef @list;
			$list[0] = $1;		# dictionary.

			s/^(.+[^ ]) +(\[.+)$/$2/ || die "$key.TAG: Unexpected format: $_";
			$list[1] = $1;		# word.

			if (s/(\]) (.+)$/$1/) {
				# reference to another word.
				$list[2] = $2;	# target.
				&write_sjis($fpwmenu, "$list[1] $_");
				&write_word_link($fpwmenu, " ", $list[2], $key, '');
			} else {
				&write_word_link($fpwmenu, '', $list[1], $key, '');
				&write_sjis($fpwmenu, " $_");
			}

			$cat = $wdic->get_file_title($list[0]);
			$cat = &reform_sjis($cat, 0);
			$cat =~ s/ //g;
			&write_sjis($fpwmenu, " s$catt\n");
		}

		if ($newline_on_death) {
			print "\n";
			$newline_on_death = 0;
		}
	}
}

##-------------------------------------------------------------------------
## PLUGIN LIST

#
# void make_plugin_list(void)
# creates a list of plugins embedded into the body. this function must be
# called after %cgraphs and %plugins have been filled.
#
sub make_plugin_list()
{
	my (%images, @files, $key, $tag);

	return if (!$opt_cgraphs && !$opt_plugin);
	print "Creating plugin list...";
	$newline_on_death = 1;

	@files = keys(%plugins) if (%plugins);
	foreach $key (keys(%cgraphs)) {
		next if (!$cgraphs{$key}{name});
		$images{$cgraphs{$key}{name}} = "PLUGIN_$key";
		push(@files, $cgraphs{$key}{name});
	}
	@files = sort @files;

	&new_menu_context('', 'PLUGIN');
	foreach $key (@files) {
		if ($plugins{$key}) {
			&write_link($fpwmenu, "$key\n", $plugins{$key}{tag});
		} elsif ($images{$key}) {
			&write_link($fpwmenu, "$key\n", $images{$key});
		}
	}

	print " done.\n";
	$newline_on_death = 0;
}

##-------------------------------------------------------------------------
## COPYRIGHT

#
# void make_copyright(void)
# creates a copyright entry using WDIC.MAN information and conversion
# date/time.
#
sub make_copyright()
{
	my ($str, $sec, $min, $hour, $day, $mon, $year);

	foreach (@copyright) {
		s/([^ ]+) +(.+$)/$2/;
		$str = $1;
		&write_sjis($fpwcopyright, $str . ' ' x (10 - length($str)) . $_);
	}

	&write_newline($fpwcopyright, 1);

	$str = "This book was generated by ";
	if (defined($ENV{'USER'})) {
		$str .= $ENV{'USER'};
	} elsif (defined($ENV{'USERNAME'})) {
		$str .= $ENV{'USERNAME'};
	} else {
		$str .= '(unknown user)';
	}

	($sec, $min, $hour, $day, $mon, $year) = localtime(time);
	$str .= sprintf(" on %04d/%02d/%02d %02d:%02d:%02d ",
			$year + 1900, $mon, $day, $hour, $min, $sec);

	$VERSION =~ /([0-9\.]+)/;
	$str .= "using fpw-wdic $1.\n";

	$fpwcopyright->add_text($str);
}

1;
